#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <intrin.h>

void __declspec(noinline) FilterComputeSymmetricFIR_8_32F_Scalar(float *__restrict dst, const float *__restrict src, size_t n, const float *__restrict kernel) {
    const float k0 = kernel[0];
    const float k1 = kernel[1];
    const float k2 = kernel[2];
    const float k3 = kernel[3];
    const float k4 = kernel[4];
    const float k5 = kernel[5];
    const float k6 = kernel[6];
    const float k7 = kernel[7];

    do {
        float v = src[7] * k0
                + (src[ 6] + src[ 8]) * k1
                + (src[ 5] + src[ 9]) * k2
                + (src[ 4] + src[10]) * k3
                + (src[ 3] + src[11]) * k4
                + (src[ 2] + src[12]) * k5
                + (src[ 1] + src[13]) * k6
                + (src[ 0] + src[14]) * k7;

        ++src;

        *dst++ = v;
    } while(--n);
}

void __declspec(noinline) __declspec(naked) FilterComputeSymmetricFIR_8_32F_Scalar_asm(float *dst, const float *src, size_t n, const float *kernel) {
    __asm {
        push	ebx
        mov		ebx, [esp+4+4]
        mov		ecx, [esp+8+4]
        mov		eax, [esp+12+4]
        mov		edx, [esp+16+4]
xloop:
        fld		dword ptr [ecx+7*4]		;x7
        fld		dword ptr [ecx+6*4]		;x7 x6
        fld		dword ptr [ecx+5*4]		;x7 x6 x5
        fxch	st(1)					;x7 x5 x6
        fadd	dword ptr [ecx+8*4]		;x7 x5 (x6+x8)
        fxch	st(2)					;(x6+x8) x5 x7
        fmul	dword ptr [edx+0*4]		;(x6+x8) x5 acc
        fld		dword ptr [ecx+4*4]		;(x6+x8) x5 acc x4
        fxch	st(2)					;(x6+x8) x4 acc x5
        fadd	dword ptr [ecx+9*4]		;(x6+x8) x4 acc (x5+x9)
        fxch	st(3)					;(x5+x9) x4 acc (x6+x8)
        fmul	dword ptr [edx+1*4]		;(x5+x9) x4 acc y68
        fld		dword ptr [ecx+3*4]		;(x5+x9) x4 acc y68 x3
        fxch	st(3)					;(x5+x9) x3 acc y68 x4
        fadd	dword ptr [ecx+10*4]	;(x5+x9) x3 acc y68 (x4+x10)
        fxch	st(4)					;(x4+x10) x3 acc y68 (x5+x9)
        fmul	dword ptr [edx+2*4]		;(x4+x10) x3 acc y68 y59
        fxch	st(1)					;(x4+x10) x3 acc y59 y68
        faddp	st(2),st				;(x4+x10) x3 acc y59
        fld		dword ptr [ecx+2*4]		;(x4+x10) x3 acc y59 x2
        fxch	st(3)					;(x4+x10) x2 acc y59 x3
        fadd	dword ptr [ecx+11*4]	;(x4+x10) x2 acc y59 (x3+x11)
        fxch	st(4)					;(x3+x11) x2 acc y59 (x4+x10)
        fmul	dword ptr [edx+3*4]		;(x3+x11) x2 acc y59 y410
        fxch	st(1)					;(x3+x11) x2 acc y410 y59
        faddp	st(2),st				;(x3+x11) x2 acc y410
        fld		dword ptr [ecx+1*4]		;(x3+x11) x2 acc y410 x1
        fxch	st(3)					;(x3+x11) x1 acc y410 x2
        fadd	dword ptr [ecx+12*4]	;(x3+x11) x1 acc y410
        fxch	st(4)					;(x2+x12) x1 acc y410 (x3+x11)
        fmul	dword ptr [edx+4*4]		;(x2+x12) x1 acc y410 y311
        fxch	st(1)					;(x2+x12) x1 acc y311 y410
        faddp	st(2),st				;(x2+x12) x1 acc y311
        fld		dword ptr [ecx+0*4]		;(x2+x12) x1 acc y311 x0
        fxch	st(3)					;(x2+x12) x0 acc y311 x1
        fadd	dword ptr [ecx+13*4]	;(x2+x12) x0 acc y311 (x1+x13)
        fxch	st(4)					;(x1+x13) x0 acc y311 (x2+x12)
        fmul	dword ptr [edx+5*4]		;(x1+x13) x0 acc y311 y212
        fxch	st(1)					;(x1+x13) x0 acc y212 y311
        faddp	st(2),st				;(x1+x13) x0 acc y212
        fxch	st(2)					;(x1+x13) y212 acc x0
        fadd	dword ptr [ecx+14*4]	;(x1+x13) y212 acc (x0+x14)
        fxch	st(3)					;(x0+x14) y212 acc (x1+x13)
        fmul	dword ptr [edx+6*4]		;(x0+x14) y212 acc y113
        fxch	st(2)					;(x0+x14) y113 acc y212
        fadd							;(x0+x14) y113 acc
        fxch	st(2)					;acc y113 (x0+x14)
        fmul	dword ptr [edx+7*4]		;acc y113 y014
        fxch	st(2)					;y014 acc y113
        fadd							;y014 acc

        fadd							;acc
        fstp	dword ptr [ebx]

        add		ebx, 4
        add		ecx, 4
        dec		eax
        jne		xloop
        pop		ebx
        ret
    }
}

void __declspec(noinline) __declspec(naked) FilterComputeSymmetricFIR_8_32F_SSE_Scalar_asm(float *dst, const float *src, size_t n, const float *kernel) {
    __asm {
        push	ebx
        mov		ebx, [esp+4+4]
        mov		ecx, [esp+8+4]
        mov		eax, [esp+12+4]
        mov		edx, [esp+16+4]
xloop:
        movss	xmm0, [ecx+7*4]

        movss	xmm1, [ecx+6*4]
        mulss	xmm0, [edx+0*4]

        movss	xmm2, [ecx+5*4]
        addss	xmm1, [ecx+8*4]

        movss	xmm3, [ecx+4*4]
        addss	xmm2, [ecx+9*4]
        mulss	xmm1, [edx+1*4]

        movss	xmm4, [ecx+3*4]
        addss	xmm3, [ecx+10*4]
        mulss	xmm2, [edx+2*4]
        addss	xmm0, xmm1

        movss	xmm5, [ecx+2*4]
        addss	xmm4, [ecx+11*4]
        mulss	xmm3, [edx+3*4]
        addss	xmm0, xmm2

        movss	xmm6, [ecx+1*4]
        addss	xmm5, [ecx+12*4]
        mulss	xmm4, [edx+4*4]
        addss	xmm0, xmm3

        movss	xmm7, [ecx+0*4]
        addss	xmm6, [ecx+13*4]
        mulss	xmm5, [edx+5*4]
        addss	xmm0, xmm4

        mulss	xmm6, [edx+6*4]
        addss	xmm0, xmm5

        addss	xmm7, [ecx+14*4]
        addss	xmm0, xmm6

        mulss	xmm7, [edx+7*4]
        add	ecx, 4

        addss	xmm0, xmm7

        movss	[ebx], xmm0
        add	ebx, 4

        dec	eax
        jne	xloop

        pop	ebx
        ret
    }
}

void __declspec(noinline) FilterComputeSymmetricFIR_8_32F_SSE(float *__restrict dst, const float *__restrict src, size_t n, const float *__restrict kernel) {
    __m128 zero = _mm_setzero_ps();
    __m128 x0 = zero;
    __m128 x1 = zero;
    __m128 x2 = zero;
    __m128 x3 = zero;
    __m128 f0;
    __m128 f1;
    __m128 f2;
    __m128 f3;

    // init filter
    __m128 k0 = _mm_loadu_ps(kernel + 0);
    __m128 k1 = _mm_loadu_ps(kernel + 4);

    f0 = _mm_shuffle_ps(k1, k1, _MM_SHUFFLE(0, 1, 2, 3));
    f1 = _mm_shuffle_ps(k0, k0, _MM_SHUFFLE(0, 1, 2, 3));
    f2 = _mm_move_ss(k0, k1);
    f2 = _mm_shuffle_ps(f2, f2, _MM_SHUFFLE(0, 3, 2, 1));
    f3 = _mm_move_ss(k1, zero);
    f3 = _mm_shuffle_ps(f3, f3, _MM_SHUFFLE(0, 3, 2, 1));

    // prime
    for(int i=0; i<14; ++i) {
        x0 = _mm_move_ss(x0, x1);
        x0 = _mm_shuffle_ps(x0, x0, _MM_SHUFFLE(0, 3, 2, 1));
        x1 = _mm_move_ss(x1, x2);
        x1 = _mm_shuffle_ps(x1, x1, _MM_SHUFFLE(0, 3, 2, 1));
        x2 = _mm_move_ss(x2, x3);
        x2 = _mm_shuffle_ps(x2, x2, _MM_SHUFFLE(0, 3, 2, 1));
        x3 = _mm_move_ss(x3, zero);
        x3 = _mm_shuffle_ps(x3, x3, _MM_SHUFFLE(0, 3, 2, 1));
//        x3 = _mm_castsi128_ps(_mm_srli_si128(_mm_castps_si128(x3), 4));

        __m128 s = _mm_load1_ps(src++);
        x0 = _mm_add_ps(x0, _mm_mul_ps(f0, s));
        x1 = _mm_add_ps(x1, _mm_mul_ps(f1, s));
        x2 = _mm_add_ps(x2, _mm_mul_ps(f2, s));
        x3 = _mm_add_ps(x3, _mm_mul_ps(f3, s));
    }

    // pipeline
    do {
        x0 = _mm_move_ss(x0, x1);
        x0 = _mm_shuffle_ps(x0, x0, _MM_SHUFFLE(0, 3, 2, 1));
        x1 = _mm_move_ss(x1, x2);
        x1 = _mm_shuffle_ps(x1, x1, _MM_SHUFFLE(0, 3, 2, 1));
        x2 = _mm_move_ss(x2, x3);
        x2 = _mm_shuffle_ps(x2, x2, _MM_SHUFFLE(0, 3, 2, 1));
        x3 = _mm_move_ss(x3, zero);
        x3 = _mm_shuffle_ps(x3, x3, _MM_SHUFFLE(0, 3, 2, 1));
//        x3 = _mm_castsi128_ps(_mm_srli_si128(_mm_castps_si128(x3), 4));

        __m128 s = _mm_load1_ps(src++);
        x0 = _mm_add_ps(x0, _mm_mul_ps(f0, s));
        x1 = _mm_add_ps(x1, _mm_mul_ps(f1, s));
        x2 = _mm_add_ps(x2, _mm_mul_ps(f2, s));
        x3 = _mm_add_ps(x3, _mm_mul_ps(f3, s));

        _mm_store_ss(dst++, x0);
    } while(--n);
}

void __declspec(naked) __cdecl FilterComputeSymmetricFIR_8_32F_SSE_asm(float *dst, const float *src, size_t n, const float *kernel) {
    __asm {
    mov	edx, esp
    and	esp, -16
    sub	esp, 80
    mov	[esp+64], edx

    mov	eax, [edx+16]
    mov	ecx, [edx+8]

    movups	xmm0, [eax+16]
    movups	xmm1, [eax]

    xorps	xmm7, xmm7
    movaps	xmm2, xmm1
    movaps	xmm3, xmm0
    shufps	xmm0, xmm0, 00011011b
    shufps	xmm1, xmm1, 00011011b
    movss	xmm2, xmm3
    shufps	xmm2, xmm2, 00111001b
    movss	xmm3, xmm7
    shufps	xmm3, xmm3, 00111001b
    movaps	[esp], xmm0
    movaps	[esp+16], xmm1
    movaps	[esp+32], xmm2
    movaps	[esp+48], xmm3

    mov	eax, 14
    xorps	xmm0, xmm0
    xorps	xmm1, xmm1
    xorps	xmm2, xmm2
    xorps	xmm3, xmm3

ploop:
    movss	xmm0, xmm1
    xorps	xmm6, xmm6
    shufps	xmm0, xmm0, 00111001b
    movss	xmm1, xmm2
    shufps	xmm1, xmm1, 00111001b
    movss	xmm2, xmm3
    shufps	xmm2, xmm2, 00111001b
    movss	xmm3, xmm6
    shufps	xmm3, xmm3, 00111001b

    movss	xmm5, [ecx]
    add	ecx, 4
    shufps	xmm5, xmm5, 0

    movaps	xmm4, xmm5
    mulps	xmm4, [esp+0]
    addps	xmm0, xmm4

    movaps	xmm4, xmm5
    mulps	xmm4, [esp+16]
    addps	xmm1, xmm4

    movaps	xmm4, xmm5
    mulps	xmm4, [esp+32]
    addps	xmm2, xmm4

    mulps	xmm5, [esp+48]
    addps	xmm3, xmm5

    dec	eax
    jne	ploop
    
    mov	eax, [edx+12]
    mov	edx, [edx+4]

    align 16
xloop:
    movss	xmm7, [ecx]
    movss	xmm0, xmm1
    shufps	xmm7, xmm7, 0

    movaps	xmm4, [esp]
    movss	xmm1, xmm2
    shufps	xmm0, xmm0, 00111001b

    movaps	xmm5, [esp+16]
    movss	xmm2, xmm3
    shufps	xmm1, xmm1, 00111001b

    xorps	xmm6, xmm6
    add	ecx, 4
    shufps	xmm2, xmm2, 00111001b

    movss	xmm3, xmm6
    movaps	xmm6, [esp+32]
    mulps	xmm4, xmm7

    shufps	xmm3, xmm3, 00111001b
    mulps	xmm5, xmm7
    mulps	xmm6, xmm7

    mulps	xmm7, [esp+48]
    addps	xmm0, xmm4
    addps	xmm1, xmm5
    addps	xmm2, xmm6
    addps	xmm3, xmm7

    movss	[edx], xmm0
    add	edx, 4

    dec	eax
    jne	xloop

    mov	esp, [esp+64]
    ret
    }
}

void __declspec(naked) __cdecl FilterComputeSymmetricFIR_8_32F_SSE2_asm(float *dst, const float *src, size_t n, const float *kernel) {
    __asm {
    mov	edx, esp
    and	esp, -16
    sub	esp, 80
    mov	[esp+64], edx

    mov	eax, [edx+16]
    mov	ecx, [edx+8]

    movups	xmm0, [eax+16]
    movups	xmm1, [eax]

    xorps	xmm7, xmm7
    movaps	xmm2, xmm1
    movaps	xmm3, xmm0
    shufps	xmm0, xmm0, 00011011b
    shufps	xmm1, xmm1, 00011011b
    movss	xmm2, xmm3
    shufps	xmm2, xmm2, 00111001b
    psrldq	xmm3, 4
    movaps	[esp], xmm0
    movaps	[esp+16], xmm1
    movaps	[esp+32], xmm2
    movaps	[esp+48], xmm3

    mov	eax, 14
    xorps	xmm0, xmm0
    xorps	xmm1, xmm1
    pxor	xmm2, xmm2
    pxor	xmm3, xmm3

ploop:
    movss	xmm0, xmm1
    movss	xmm1, xmm2
    movss	xmm2, xmm3
    shufps	xmm0, xmm0, 00111001b
    shufps	xmm1, xmm1, 00111001b
    shufps	xmm2, xmm2, 00111001b
    psrldq	xmm3, 4

    movss	xmm5, [ecx]
    add	ecx, 4
    shufps	xmm5, xmm5, 0

    movaps	xmm4, xmm5
    mulps	xmm4, [esp+0]
    addps	xmm0, xmm4

    movaps	xmm4, xmm5
    mulps	xmm4, [esp+16]
    addps	xmm1, xmm4

    movaps	xmm4, xmm5
    mulps	xmm4, [esp+32]
    addps	xmm2, xmm4

    mulps	xmm5, [esp+48]
    addps	xmm3, xmm5

    dec	eax
    jne	ploop
    
    mov	eax, [edx+12]
    mov	edx, [edx+4]
    movaps	xmm4, [esp]
    movaps	xmm5, [esp+16]
    movaps	xmm6, [esp+32]

    align 16
xloop:
    movss	xmm7, [ecx]
    movss	xmm0, xmm1

    shufps	xmm7, xmm7, 0

    mulps	xmm4, xmm7
    movss	xmm1, xmm2

    shufps	xmm0, xmm0, 00111001b
    add		ecx, 4

    mulps	xmm5, xmm7
    movss	xmm2, xmm3

    shufps	xmm1, xmm1, 00111001b
    movaps	xmm6, xmm7

    mulps	xmm6, [esp+32]

    shufps	xmm2, xmm2, 00111001b
    addps	xmm0, xmm4

    mulps	xmm7, [esp+48]
    add		edx, 4

    psrldq	xmm3, 4
    addps	xmm1, xmm5

    movaps	xmm4, [esp]
    dec		eax

    movaps	xmm5, [esp+16]
    addps	xmm2, xmm6

    movss	[edx-4], xmm0

    addps	xmm3, xmm7
    jne		xloop

    mov	esp, [esp+64]
    ret
    }
}

void __declspec(naked) __cdecl FilterComputeSymmetricFIR_8_32F_SSSE3_asm(float *dst, const float *src, size_t n, const float *kernel) {
    __asm {
    push ebx
    push edi
    mov	edx, esp
    and	esp, -16
    sub	esp, 80
    mov	[esp+64], edx

    mov	eax, [edx+24]
    mov	ecx, [edx+16]

    movups	xmm0, [eax+16]
    movups	xmm1, [eax]

    ;    7  6  5  4  3  2  1  0  1  2  3  4  5  6  7
    ;xmm03|02|01|00|13|12|11|10|23|22|21|20|33|32|31|30
    movaps	xmm2, xmm1
    movaps	xmm3, xmm0
    shufps	xmm2, xmm2, 00011011b
    shufps	xmm3, xmm3, 00011011b
    palignr	xmm2, xmm3, 12
    pslldq	xmm3, 4

    movaps	[esp], xmm0
    movaps	[esp+16], xmm1
    movaps	[esp+32], xmm2
    movaps	[esp+48], xmm3

    mov	eax, 14
    pxor	xmm0, xmm0
    pxor	xmm1, xmm1
    pxor	xmm2, xmm2
    pxor	xmm3, xmm3

ploop:
    palignr	xmm0, xmm1, 12
    palignr	xmm1, xmm2, 12
    palignr	xmm2, xmm3, 12
    pslldq	xmm3, 4

    movss	xmm5, [ecx]
    add	ecx, 4
    shufps	xmm5, xmm5, 0

    movaps	xmm4, xmm5
    mulps	xmm4, [esp+0]
    addps	xmm0, xmm4

    movaps	xmm4, xmm5
    mulps	xmm4, [esp+16]
    addps	xmm1, xmm4

    movaps	xmm4, xmm5
    mulps	xmm4, [esp+32]
    addps	xmm2, xmm4

    mulps	xmm5, [esp+48]
    addps	xmm3, xmm5

    dec	eax
    jne	ploop
    
    mov	eax, [edx+20]
    mov	edx, [edx+12]

    movaps	xmm4, [esp]
    movaps	xmm5, [esp+16]
    mov		ebx, esp
    lea		edi, [esp+48]

    align 16
xloop:
    ;0
    movss	xmm7, [ecx]
    add		ecx, 4

    ;1
    shufps	xmm7, xmm7, 0
    add		edx, 4

    ;2 -> 7
    mulps	xmm4, xmm7

    ;3

    ;4 -> 9
    mulps	xmm5, xmm7

    ;5
    palignr	xmm0, xmm1, 12
    movaps	xmm6, xmm7

    ;6 -> 11
    mulps	xmm6, [ebx+32]

    ;7 -> 12
    palignr	xmm1, xmm2, 12
    addps	xmm0, xmm4

    ;8 -> 13
    mulps	xmm7, [edi]

    ;9 -> 14
    palignr	xmm2, xmm3, 12
    addps	xmm1, xmm5

    ;10
    pslldq	xmm3, 4
    dec		eax

    ;11 -> 16(1)
    movaps	xmm4, [ebx]
    addps	xmm2, xmm6

    ;12
    movaps	xmm5, [ebx+16]
    movaps	xmm6, xmm0

    ;13 -> 18(3)
    shufps	xmm6, xmm6, 255
    addps	xmm3, xmm7
    
    ;14
    movss	[edx-4], xmm6
    jne		xloop

    mov	esp, [esp+64]
    pop	edi
    pop	ebx
    ret
    }
}

float k[15];
float s[4096 + 14];
float d1[4096];
float d2[4096];
float d3[4096];
float d4[4096];
float d5[4096];
float d6[4096];
float d7[4096];

int main() {
    __int64 b1 = 0x7fffffffffffffff;
    __int64 b2 = 0x7fffffffffffffff;
    __int64 b3 = 0x7fffffffffffffff;
    __int64 b4 = 0x7fffffffffffffff;
    __int64 b5 = 0x7fffffffffffffff;
    __int64 b6 = 0x7fffffffffffffff;
    __int64 b7 = 0x7fffffffffffffff;

    for(int i=0; i<15; ++i)
        k[i] = (float)rand() / (float)RAND_MAX;

    for(int i=0; i<4096+14; ++i)
        s[i] = (float)rand() / (float)RAND_MAX;

    for(int i=0; i<10; ++i) {
        __int64 t1 = __rdtsc();
        FilterComputeSymmetricFIR_8_32F_Scalar(d1, s, 4096, k);
        __int64 dt = __rdtsc() - t1;

        if (b1 > dt)
            b1 = dt;
    }

    for(int i=0; i<10; ++i) {
        __int64 t1 = __rdtsc();
        FilterComputeSymmetricFIR_8_32F_SSE(d2, s, 4096, k);
        __int64 dt = __rdtsc() - t1;

        if (b2 > dt)
            b2 = dt;
    }

    for(int i=0; i<10; ++i) {
        __int64 t1 = __rdtsc();
        FilterComputeSymmetricFIR_8_32F_SSE_asm(d3, s, 4096, k);
        __int64 dt = __rdtsc() - t1;

        if (b3 > dt)
            b3 = dt;
    }

    for(int i=0; i<10; ++i) {
        __int64 t1 = __rdtsc();
        FilterComputeSymmetricFIR_8_32F_SSSE3_asm(d4, s, 4096, k);
        __int64 dt = __rdtsc() - t1;

        if (b4 > dt)
            b4 = dt;
    }

    for(int i=0; i<10; ++i) {
        __int64 t1 = __rdtsc();
        FilterComputeSymmetricFIR_8_32F_SSE_Scalar_asm(d5, s, 4096, k);
        __int64 dt = __rdtsc() - t1;

        if (b5 > dt)
            b5 = dt;
    }

    for(int i=0; i<10; ++i) {
        __int64 t1 = __rdtsc();
        FilterComputeSymmetricFIR_8_32F_Scalar_asm(d6, s, 4096, k);
        __int64 dt = __rdtsc() - t1;

        if (b6 > dt)
            b6 = dt;
    }

    for(int i=0; i<10; ++i) {
        __int64 t1 = __rdtsc();
        FilterComputeSymmetricFIR_8_32F_SSE2_asm(d7, s, 4096, k);
        __int64 dt = __rdtsc() - t1;

        if (b7 > dt)
            b7 = dt;
    }

    printf("%lld %lld %lld %lld %lld %lld %lld\n", b1, b2, b3, b4, b5, b6, b7);

    for(int i=0; i<4096; ++i) {
        float x = d1[i];
        float y = d2[i];
        float z = d3[i];
        float w = d4[i];
        float s = d5[i];
        float t = d6[i];
        float u = d7[i];

        if (fabsf(x - y) + fabsf(y - z) + fabsf(z - w) + fabsf(w - s) + fabsf(s - t) + fabsf(t - u) > 1e-5f) {
            printf("FAIL[%d]: %g %g %g %g %g %g %g\n", i, x, y, z, w, s, t, u);
            break;
        }
    }
}
